--- old/usr/src/uts/common/fs/zfs/dmu_send.c	Mon Oct 29 18:19:33 2012
+++ new/usr/src/uts/common/fs/zfs/dmu_send.c	Mon Oct 29 18:19:33 2012
@@ -51,6 +51,7 @@
 int zfs_send_corrupt_data = B_FALSE;
 
 static char *dmu_recv_tag = "dmu_recv_tag";
+char *tmp_dmu_recv_tag = "tmp_dmu_recv_tag";
 
 static int
 dump_bytes(dmu_sendarg_t *dsp, void *buf, int len)
@@ -643,6 +644,22 @@
 			return (ENODEV);
 	}
 
+	/*
+	 * Check dataset and snapshot quotas before receiving. We'll recheck
+	 * again at the end, but might as well abort before receiving if we're
+	 * already over quota.
+	 */
+	if (dd->dd_parent != NULL) {
+		err = dsl_dir_dscount_check(dd->dd_parent, NULL, 1, NULL);
+		if (err != 0)
+			return (err);
+	}
+
+	err = dsl_snapcount_check(dd, tx, 1, NULL);
+	if (err != 0)
+		return (err);
+
+
 	return (0);
 }
 
@@ -725,6 +742,11 @@
 		/* if full, most recent snapshot must be $ORIGIN */
 		if (ds->ds_phys->ds_prev_snap_txg >= TXG_INITIAL)
 			return (ENODEV);
+
+		/* Check snapshot quota before receiving */
+		err = dsl_snapcount_check(ds->ds_dir, tx, 1, NULL);
+		if (err != 0)
+			return (err);
 	}
 
 	/* temporary clone name must not exist */
@@ -1547,6 +1569,7 @@
 	char *tosnap;
 	uint64_t creation_time;
 	uint64_t toguid;
+	boolean_t is_new;
 };
 
 static int
@@ -1555,7 +1578,21 @@
 	dsl_dataset_t *ds = arg1;
 	struct recvendsyncarg *resa = arg2;
 
-	return (dsl_dataset_snapshot_check(ds, resa->tosnap, tx));
+	if (resa->is_new) {
+		/* re-check the dataset quota now that recv is complete */
+		dsl_dir_t *dd;
+		int err;
+
+		dd = ds->ds_dir;
+		if (dd->dd_parent != NULL) {
+			err = dsl_dir_dscount_check(dd->dd_parent, NULL, 1,
+			    NULL);
+			if (err != 0)
+				return (err);
+		}
+	}
+
+	return (dsl_dataset_snapshot_check(ds, resa->tosnap, 1, tx));
 }
 
 static void
@@ -1564,6 +1601,11 @@
 	dsl_dataset_t *ds = arg1;
 	struct recvendsyncarg *resa = arg2;
 
+	if (resa->is_new)
+		/* update the dataset counts */
+		dsl_dir_dscount_adjust(ds->ds_dir->dd_parent, tx, 1, B_FALSE,
+		    B_TRUE);
+
 	dsl_dataset_snapshot_sync(ds, resa->tosnap, tx);
 
 	/* set snapshot's creation time and guid */
@@ -1616,7 +1658,8 @@
 	} else {
 		mutex_exit(&ds->ds_recvlock);
 		dsl_dataset_rele(ds, dmu_recv_tag);
-		(void) dsl_dataset_destroy(drc->drc_real_ds, dmu_recv_tag,
+		/* tag indicates temporary ds to dsl_dir_destroy_sync */
+		(void) dsl_dataset_destroy(drc->drc_real_ds, tmp_dmu_recv_tag,
 		    B_FALSE);
 		return (EBUSY);
 	}
@@ -1624,6 +1667,7 @@
 	resa.creation_time = drc->drc_drrb->drr_creation_time;
 	resa.toguid = drc->drc_drrb->drr_toguid;
 	resa.tosnap = drc->drc_tosnap;
+	resa.is_new = B_FALSE;
 
 	err = dsl_sync_task_do(ds->ds_dir->dd_pool,
 	    recv_end_check, recv_end_sync, ds, &resa, 3);
@@ -1637,7 +1681,9 @@
 	if (err == 0 && drc->drc_guid_to_ds_map != NULL)
 		(void) add_ds_to_guidmap(drc->drc_guid_to_ds_map, ds);
 	dsl_dataset_disown(ds, dmu_recv_tag);
-	myerr = dsl_dataset_destroy(drc->drc_real_ds, dmu_recv_tag, B_FALSE);
+	/* tag indicates temporary ds to dsl_dir_destroy_sync */
+	myerr = dsl_dataset_destroy(drc->drc_real_ds, tmp_dmu_recv_tag,
+	    B_FALSE);
 	ASSERT0(myerr);
 	return (err);
 }
@@ -1659,6 +1705,7 @@
 	resa.creation_time = drc->drc_drrb->drr_creation_time;
 	resa.toguid = drc->drc_drrb->drr_toguid;
 	resa.tosnap = drc->drc_tosnap;
+	resa.is_new = B_TRUE;
 
 	err = dsl_sync_task_do(ds->ds_dir->dd_pool,
 	    recv_end_check, recv_end_sync, ds, &resa, 3);
